/**
 * @file prt_deque.c
 * @brief double-ended queue
 * @author Jakub Kurzak
 *
 * Thread-safe and size-aware.
 * Built on top of icl_list.
 * Uses Pthread spinlocks.
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#include "icl_deque.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief deque constructor
 *
 * @return new deque. NULL on error.
 */
icl_deque_t* icl_deque_new()
{
    icl_deque_t *deque = (icl_deque_t*)malloc(sizeof(icl_deque_t));
    if (deque == NULL)
        return NULL;

    deque->list = icl_list_new();
    if (deque->list == NULL) {
        free(deque);
        return NULL;
    }
    if (pthread_spin_init(&deque->spinlock, PTHREAD_PROCESS_PRIVATE) != 0) {
        free(deque->list);
        free(deque);
        return NULL;
    }
    deque->size = 0;
    return deque;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief deque destructor
 *
 * @param deque
 * @param free_func function that frees the node's data
 *
 * @retval  0 success
 * @retval -1 failure
 */
int icl_deque_destroy(icl_deque_t *deque, void (*free_func)(void*))
{
    if (deque == NULL)
        return -1;

    if (pthread_spin_destroy(&deque->spinlock) != 0)
        return -1;

    if (icl_list_destroy(deque->list, free_func) != 0)
        return -1;

    free(deque);
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Get the first node in the deque.
 *
 * @param deque
 *
 * @return the node at the front of the deque. NULL if empty or error.
 */
icl_node_t* icl_deque_first(icl_deque_t *deque)
{
    if (deque == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_first(deque->list);
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Get the next node in the deque.
 *
 * @param deque
 * @param node
 *
 * @return next node. NULL if empty or error.
 */
icl_node_t* icl_deque_next(icl_deque_t *deque, icl_node_t *node)
{
    if (deque == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *next = icl_list_next(deque->list, node);
    pthread_spin_unlock(&deque->spinlock);

    return next;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Insert the node at the end of the deque.
 *
 * @param deque
 * @param data
 *
 * @return the new node. NULL on error.
 */
icl_node_t* icl_deque_append(icl_deque_t *deque, void *data)
{
    if (deque == NULL || data == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_append(deque->list, data);
    deque->size++;
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Insert the node at the front of the deque.
 *
 * @param deque
 * @param data
 *
 * @return the new node. NULL on error.
 */
icl_node_t* icl_deque_prepend(icl_deque_t *deque, void *data)
{
    if (deque == NULL || data == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_prepend(deque->list, data);
    deque->size++;
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Delete the node from the deque.
 *
 * @param deque
 * @param node
 * @param free_func function that frees the node's data
 *
 * @retval  0 success
 * @retval -1 failure
 */
int icl_deque_delete(icl_deque_t *deque, icl_node_t *node, void (*free_func)(void*))
{
    if (deque == NULL || node == NULL)
        return -1;

    pthread_spin_lock(&deque->spinlock);
    if (icl_list_delete(deque->list, node, free_func) == 0) {
        pthread_spin_unlock(&deque->spinlock);
        deque->size--;
        return 0;
    }
    else {
        pthread_spin_unlock(&deque->spinlock);
        return -1;
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Return the deque size.
 *
 * @param deque
 *
 * @return size of the deque. -1 on error.
 */
int icl_deque_size(icl_deque_t *deque)
{
    if (deque == NULL)
        return -1;

    pthread_spin_lock(&deque->spinlock);
    int size = deque->size;
    pthread_spin_unlock(&deque->spinlock);
    return size;
}
