/**
 * @file prt_channel.h
 * @brief PRT data channel
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#ifndef PRT_CHANNEL_H
#define PRT_CHANNEL_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_packet_s;

/**
 * @brief VDP's data channel
 * Implements a data link between a pair of VDPs.
 * Identifies the source and destination VDPs by tuples.
 * Contains a list of data packets, protected by a spinlock.
 */
typedef struct prt_channel_s {
    struct prt_vdp_s *vdp;          // back pointer to the parent VDP
    int count;                      // number of elements in a packet
    MPI_Datatype datatype;          // datatype of elements in a packet
    int *src_tuple;                 // tuple of the source VDP
    int src_slot;                   // slot in the source VDP
    int *dst_tuple;                 // tuple of the destination VDP
    int dst_slot;                   // slot in the destination VDP
    int src_node;                   // source node
    int dst_node;                   // destination node
    int tag;                        // tag for process-to-process communication
    icl_deque_t *packets;           // list of data packets
} prt_channel_t;

/**
 * @brief VDP's data channel direction
 * Identifies the direction of a VDP channel during insertion.
 */
typedef enum prt_channel_direction_e {
    PrtInputChannel, PrtOutputChannel
} prt_channel_direction_t;

prt_channel_t* prt_channel_new(
    int count, MPI_Datatype datatype,
    int *src_tuple, int src_slot,
    int *dst_tuple, int dst_slot);
void prt_channel_delete(prt_channel_t *channel);
void prt_channel_push(prt_channel_t *channel, struct prt_packet_s *packet);
struct prt_packet_s *prt_channel_pop(prt_channel_t *channel);
int prt_channel_empty(prt_channel_t *channel);
int prt_channel_compare(void *channel1, void *channel2);

#ifdef __cplusplus
}
#endif

#endif /* PRT_CHANNEL_H */
