/**
 * @file prt_request.c
 * @brief PRT communication request implementation
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#include "prt_request.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief request constructor
 *
 * @param packet
 * @param count
 * @param datatype
 * @param peer
 * @param tag
 *
 * @return new request
 */
prt_request_t *prt_request_new(
    prt_packet_t *packet, int count,
    MPI_Datatype datatype, int peer, int tag)
{
    prt_request_t *request = (prt_request_t*)malloc(sizeof(prt_request_t));
    prt_assert(request != NULL, "malloc failed");
    request->packet = packet;
    request->count = count;
    request->datatype = datatype;
    request->peer = peer;
    request->tag = tag;
    return request;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief request destructor
 * Request is only an envelope for a packet.
 * Request destruction does not affect the packet.
 * Need be, packet is destroyed separately (e.g. send completion).
 *
 * @param request
 */
void prt_request_destroy(prt_request_t *request)
{
    // Free the request.
    free(request);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Post a send request.
 * Detects a possible overflow of the request size.
 *
 * @param request
 */
void prt_request_send(prt_request_t *request)
{
    int retval;
    int type_size;
    long request_size;

    // Compute type size. Check for overflow.
    MPI_Type_size(request->datatype, &type_size);
    request_size = (long)request->count*type_size;
    prt_assert(request_size <= INT_MAX, "request size overflow");

    // Post and trace the request. Check the status.
    svg_trace_start(0);
    retval = MPI_Isend(
        request->packet->data,
        request->count*type_size,
        MPI_BYTE,
        request->peer,
        request->tag,
        MPI_COMM_WORLD,
        &request->request);
    svg_trace_stop(0, -DarkKhaki);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Isend");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Post a receive request.
 * Detects a possible overflow of the request size.
 *
 * @param request
 */
void prt_request_recv(prt_request_t *request)
{
    int retval;
    int type_size;
    long request_size;

    // Compute type size. Check for overflow.
    MPI_Type_size(request->datatype, &type_size);
    request_size = (long)request->count*type_size;
    prt_assert(request_size <= INT_MAX, "request size overflow");

    // Post and trace the request. Check the status.
    svg_trace_start(0);
    retval = MPI_Irecv(
        request->packet->data,
        request->count*type_size,
        MPI_BYTE,
        request->peer,
        request->tag,
        MPI_COMM_WORLD,
        &request->request);
    svg_trace_stop(0, -OliveDrab);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Isend");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Test a request.
 * Trace only completed requests.
 *
 * @param request
 *
 * @retval 0 Operation pending.
 * @retval 1 Operation completed.
 */
int prt_request_test(prt_request_t *request)
{
    int flag;
    int retval;

    // Test the request.
    // Check the status.
    // Trace if completed.
    svg_trace_start(0);
    retval = MPI_Test(&request->request, &flag, &request->status);
    if (flag == 1)
        svg_trace_stop(0, -RosyBrown);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Test");
    return flag;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Cancel a request.
 * Cancel the request, release the packet, free the request object.
 *
 * @param request
 */
void prt_request_cancel(prt_request_t *request)
{
    int retval;
    // Cancel the request.
    retval = MPI_Cancel(&request->request);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Cancel");

    // Release the packet.
    prt_packet_release(request->packet);

    // Free the request.
    free(request);
}
