/**
 * @file prt_tuple.c
 * @brief PRT tuple implementation
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#include "prt_tuple.h"

static unsigned int fnv_hash(void *key, int len);

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple constructor
 * Allocates memory for the tuple plus the termination symbol (INT_MAX).
 *
 * @param len length of the tuple
 * @param ... list of elements of type int
 *
 * @return pointer to an integer array terminated with INT_MAX
 */
int *prt_tuple_new(int len, ...)
{
    int *tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(tuple != NULL, "malloc failed");
    int i;
    va_list ap;
    va_start(ap, len);
    for (i = 0; i < len; i++)
        tuple[i] = va_arg(ap, int);
    va_end(ap);
    tuple[len] = INT_MAX;
    return tuple;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple length
 *
 * @param tuple
 *
 * @return tuple length without the terminating symbol
 */
int prt_tuple_len(int *tuple)
{
    int *p = tuple;
    while (*p++ != INT_MAX);
    return (p-tuple-1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple concatenate
 * Join a variable length, NULL-terminated, list of tuples.
 *
 * @param first_tuple
 * @param ... the other tuples
 *
 * @return the combined tuple
 */
int* prt_tuple_cat(int *first_tuple, ...)
{
    int len;
    int *tuple;
    va_list ap;
    va_start(ap, first_tuple);
    len = prt_tuple_len(first_tuple);
    while ((tuple = va_arg(ap, int*)) != NULL)
        len += prt_tuple_len(tuple);
    va_end(ap);

    int *out_tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(out_tuple != NULL, "malloc failed");
    va_start(ap, first_tuple);
    while(*first_tuple != INT_MAX)
        *out_tuple++ = *first_tuple++;
    while ((tuple = va_arg(ap, int*)) != NULL)
        while(*tuple != INT_MAX)
            *out_tuple++ = *tuple++;
    va_end(ap);

    *out_tuple = INT_MAX;
    return (out_tuple-len);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple destructor
 *
 * @param tuple
 */
void prt_tuple_delete(int *tuple)
{
    // Free the tuple.
    free(tuple);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple copy constructor
 *
 * @param in_tuple
 *
 * @return new copy
 */
int *prt_tuple_copy(int *in_tuple)
{
    int i;
    int len = 0;
    int *out_tuple;
    while (in_tuple[len] != INT_MAX)
        len++;
    out_tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(out_tuple != NULL, "malloc failed");
    for (i = 0; i < len; i++)
        out_tuple[i] = in_tuple[i];
    out_tuple[len] = INT_MAX;
    return out_tuple;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple comparison
 *
 * @param tuple_a
 * @param tuple_b
 *
 * @retval -1 tuple_a is less than tuple_b.
 * @retval  0 tuple_a is equal to tuple_b.
 * @retval  1 tuple_a is greater than tuple_b.
 */
int prt_tuple_compare(void *tuple_a, void *tuple_b)
{
    int *tupa = (int*)tuple_a;
    int *tupb = (int*)tuple_b;
    while (*tupa != INT_MAX && *tupb != INT_MAX) {
        if (*tupa < *tupb) return -1;
        else if (*tupa > *tupb) return 1;
        tupa++;
        tupb++;
    }
    if (*tupa == INT_MAX && *tupb == INT_MAX) return 0;
    else if (*tupa == INT_MAX) return -1;
    else if (*tupb == INT_MAX) return 1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple equality check
 * Check if tuples are identical in length and content.
 *
 * @param tuple_a
 * @param tuple_b
 *
 * @retval 0 Tuples differ.
 * @retval 1 Tuples are identical.
 */
int prt_tuple_equal(void *tuple_a, void *tuple_b)
{
    return (prt_tuple_compare(tuple_a, tuple_b) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple print
 *
 * @param tuple
 */
void prt_tuple_print(int *tuple)
{
    while (*tuple != INT_MAX)
        printf("%3d", *tuple++);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief tuple hash
 * Required by the VSA's tuples hash table.
 * Computes the lenght in characters and calls a string hash function.
 *
 * @param tuple
 *
 * @return hash
 */
unsigned int prt_tuple_hash(void *tuple)
{
    int len = 0;
    int *tup = (int*)tuple;
    while (tup[len] != INT_MAX)
        len++;
    return fnv_hash(tuple, len*sizeof(int));
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Fowler/Noll/Vo hash
 * http://eternallyconfuzzled.com/tuts/algorithms/jsw_tut_hashing.aspx
 *
 * @param key pointer to the string
 * @param len length of the string
 *
 * @return hash
 */
static unsigned int fnv_hash(void *key, int len)
{
    unsigned char *p = key;
    unsigned h = 2166136261;
    int i;
    for (i = 0; i < len; i++)
        h = (h * 16777619) ^ p[i];
    return h;
}
