/**
 * @file prt_vdp.h
 * @brief Virtual Data Processor (VDP)
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#ifndef PRT_VDP_H
#define PRT_VDP_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_thread_s;
struct prt_channel_s;
enum prt_channel_direction_e;

/**
 * @brief VDP's function pointer
 * Defines the type of the pointer to the VDP's function.
 */
typedef void (*prt_vdp_function_t)(int*, int, struct prt_channel_s**, struct prt_channel_s**, void*, void*);

/**
 * @brief Virtual Data Processor (VDP)
 * Is uniquely identified by a tuple.
 * Fires for a predefined number of cycles.
 * Has a fixed number of input and output channels.
 * Has a persistent local store.
 */
typedef struct prt_vdp_s {
    struct prt_thread_s *thread;    // back pointer to the parent thread
    int *tuple;                     // the VDP's tuple
    int counter;                    // the VDP's countdown counter
    int num_inputs;                 // number of input channels
    struct prt_channel_s **input;   // array of input channels
    int num_outputs;                // number of output channels
    struct prt_channel_s **output;  // array of output channels
    prt_vdp_function_t function;    // VDP's function
    void *local_store;              // VDP's persistent local storage
    int color;                      // RGB color for tracing
} prt_vdp_t;

prt_vdp_t* prt_vdp_new(
    int *tuple, int counter,
    prt_vdp_function_t function,
    size_t local_store_size,
    int num_inputs, int num_outputs, int color);
void prt_vdp_delete(prt_vdp_t *vdp);
void prt_vdp_annihilate(prt_vdp_t *vdp);
void prt_vdp_channel_insert(
    prt_vdp_t *vdp, struct prt_channel_s *channel,
    enum prt_channel_direction_e direction, int slot);
int prt_vdp_ready(prt_vdp_t *vdp);

#ifdef __cplusplus
}
#endif

#endif /* PRT_VDP_H */
