/**
 * @file prt_vsa.c
 * @brief Virtual Systolic Array (VSA) implementation
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#include "prt_vsa.h"

extern int prt_tuple_equal(void *tuple_a, void *tuple_b);
extern unsigned int prt_tuple_hash(void *tuple);

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API
 * @brief VSA constructor
 *
 * @param num_threads number of local worker threads
 * @param global_store VSA's global store accessible to all VDPs
 * @param vdp_to_core function mapping VDP's tuple to a global core number
 * @param max_vdps maximum number of VDP's within a node
 *
 * @return new VSA
 */
prt_vsa_t* prt_vsa_new(
    int num_threads, void *global_store, int (*vdp_to_core)(int*, void*, int))
{
    // Allocate the VSA.
    prt_vsa_t *vsa = (prt_vsa_t*)malloc(sizeof(prt_vsa_t));
    prt_assert(vsa != NULL, "malloc failed");

    // Init the VSA.
    MPI_Comm_rank(MPI_COMM_WORLD, &vsa->node_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &vsa->num_nodes);
    vsa->num_threads = num_threads;
    vsa->num_cores = vsa->num_nodes*vsa->num_threads;
    vsa->global_store = global_store;
    vsa->vdp_to_core = vdp_to_core;

    // Init config & proxy.
    vsa->config = prt_config_new();
    vsa->proxy = prt_proxy_new(num_threads);
    vsa->proxy->vsa = vsa;

    // Init pthreads.
    pthread_attr_init(&vsa->thread_attr);
    pthread_attr_setscope(&vsa->thread_attr, PTHREAD_SCOPE_SYSTEM);
    pthread_setconcurrency(vsa->num_threads+1);

    int i;
    // Initialize the threads.
    vsa->thread = (prt_thread_t**)malloc(vsa->num_threads*sizeof(prt_thread_t*));
    prt_assert(vsa->thread != NULL, "malloc failed");
    for (i = 0; i < vsa->num_threads; i++) {
        vsa->thread[i] = prt_thread_new(i, vsa->node_rank*vsa->num_threads+i);
        vsa->thread[i]->vsa = vsa;
    }

    // Initialize the VDPs hash.
    int nbuckets = PRT_VSA_MAX_VDPS_PER_NODE;
    vsa->vdps_hash = icl_hash_create(nbuckets, prt_tuple_hash, prt_tuple_equal);

    // Allocate the array of channel lists.
    vsa->channel_lists = (icl_list_t**)calloc(vsa->num_nodes, sizeof(icl_list_t*));
    prt_assert(vsa->channel_lists != NULL, "malloc failed");

    // Return the VSA.
    return vsa;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API
 * @brief VSA destructor
 *
 * @param VSA
 */
void prt_vsa_delete(prt_vsa_t *vsa)
{
    // Destroy the VDPs hash.
    icl_hash_destroy(vsa->vdps_hash, NULL, (void(*)(void*))prt_vdp_delete);

    // Delete config & proxy.
    prt_config_delete(vsa->config);
    prt_proxy_delete(vsa->proxy);

    int i;
    // Pthreads cleanup.
    pthread_attr_destroy(&vsa->thread_attr);
    for (i = 0; i < vsa->num_threads; i++)
        prt_thread_delete(vsa->thread[i]);
    free(vsa->thread);

    // Free the VSA.
    free(vsa);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API
 * @brief Inserts a new VDP into a VSA.
 * Destroys VDPs that do not belong to this node.
 * Puts the VDP in the list of VDPs of the owner thread.
 * Connects corresponding input and output channels of intra-node VDPs.
 * Builds the list of channel connections to other nodes.
 *
 * @param vsa
 * @param vdp
 */
void prt_vsa_vdp_insert(prt_vsa_t *vsa, prt_vdp_t *vdp)
{
    // Check arguments.
    prt_assert(vsa != NULL, "NULL VSA");
    prt_assert(vdp != NULL, "NULL VDP");

    // Compute global core number and process rank.
    int core = vsa->vdp_to_core(vdp->tuple, vsa->global_store, vsa->num_cores);
    int node_rank = core / vsa->num_threads;
    int thread_rank = core % vsa->num_threads;

    // IF VDP not in this node.
    if (node_rank != vsa->node_rank) {
        // Destroy along with all channels and return.
        prt_vdp_annihilate(vdp);
        return;
    }
    // Insert in the VSA's VDP hash.
    icl_entry_t *entry = icl_hash_insert(
        vsa->vdps_hash, (void*)vdp->tuple, (void*)vdp);
    prt_assert(entry != NULL, "icl_hash_insert failed");

    // Insert in the thread's VDPs list.
    icl_list_t *node = icl_list_append(vsa->thread[thread_rank]->vdps, vdp);
    prt_assert(node != NULL, "icl_list_append failed");
    vdp->thread = vsa->thread[thread_rank];

    // Merge intra-node channels.
    prt_vsa_vdp_merge_channels(vsa, vdp, node_rank);
    // Track channel tags for inter-node communication.
    prt_vsa_vdp_track_tags(vsa, vdp, core, node_rank);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Connects corresponding input and output channels of intra-node VDPs.
 *
 * @param vsa
 * @param vdp
 */
void prt_vsa_vdp_merge_channels(prt_vsa_t *vsa, prt_vdp_t *vdp, int node_rank)
{
    int i;
    // FOR each input channel.
    for (i = 0; i < vdp->num_inputs; i++) {
        prt_channel_t *channel = vdp->input[i];
        if (channel != NULL) {
            // Look for maximum channel size.
            prt_proxy_max_packet_size(vsa->proxy, channel);
            // Look up the source VDP.
            prt_vdp_t *src_vdp =
                icl_hash_find(vsa->vdps_hash, (void*)channel->src_tuple);
            // IF source VDP found.
            if (src_vdp != NULL) {
                // Check for channel tuple mismatch.
                int *src_vdp_dst_tuple =
                    src_vdp->output[channel->src_slot]->dst_tuple;
                prt_assert(prt_tuple_equal(src_vdp_dst_tuple, vdp->tuple),
                    "VDP channel tuple mismatch");
                // Swap this channel to the existing channel.
                vdp->input[i] = src_vdp->output[channel->src_slot];
                prt_channel_delete(channel);
            }
        }
    }
    // FOR each output channel.
    for (i = 0; i < vdp->num_outputs; i++) {
        prt_channel_t *channel = vdp->output[i];
        if (channel != NULL) {
            // Look for maximum channel size.
            prt_proxy_max_packet_size(vsa->proxy, channel);
            // Look up the destination VDP.
            prt_vdp_t *dst_vdp =
                icl_hash_find(vsa->vdps_hash, (void*)channel->dst_tuple);
            // IF destination VDP found.
            if (dst_vdp != NULL) {
                // Check for channel tuple mismatch.
                int *dst_vdp_src_tuple =
                    dst_vdp->input[channel->dst_slot]->src_tuple;
                prt_assert(prt_tuple_equal(dst_vdp_src_tuple, vdp->tuple),
                    "VDP channel tuple mismatch");
                // Swap this channel for the existing channel.
                vdp->output[i] = dst_vdp->input[channel->dst_slot];
                prt_channel_delete(channel);
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Builds the list of channel connections to other nodes.
 *
 * @param vsa
 * @param vdp
 * @param core VDP's core number
 * @param node_rank VDP's node rank
 */
void prt_vsa_vdp_track_tags(
    prt_vsa_t *vsa, prt_vdp_t *vdp, int core, int node_rank)
{
    int i;
    // FOR each input channel.
    for (i = 0; i < vdp->num_inputs; i++) {
        prt_channel_t *channel = vdp->input[i];
        if (channel != NULL) {
            channel->dst_node = node_rank;
            int src_core = vsa->vdp_to_core(
                channel->src_tuple, vsa->global_store, vsa->num_cores);
            int src_node = src_core / vsa->num_threads;
            channel->src_node = src_node;
            // IF another node is the source.
            if (src_node != vsa->node_rank) {
                // Create the list if empty.
                if (vsa->channel_lists[src_node] == NULL) {
                    vsa->channel_lists[src_node] = icl_list_new();
                    prt_assert(vsa->channel_lists[src_node] != NULL,
                        "icl_list_new failed");
                }
                // Add the channel to the list.
                icl_list_t *node = icl_list_isort(
                    vsa->channel_lists[src_node], channel, prt_channel_compare);
                prt_assert(node != NULL, "icl_list_isort failed");
            }
        }
    }
    // FOR each output channel.
    for (i = 0; i < vdp->num_outputs; i++) {
        prt_channel_t *channel = vdp->output[i];
        if (channel != NULL) {
            channel->src_node = node_rank;
            int dst_core = vsa->vdp_to_core(
                channel->dst_tuple, vsa->global_store, vsa->num_cores);
            int dst_node = dst_core / vsa->num_threads;
            channel->dst_node = dst_node;
            // IF another node is the destination.
            if (dst_node != vsa->node_rank) {
                // Create the list if empty.
                if (vsa->channel_lists[dst_node] == NULL) {
                    vsa->channel_lists[dst_node] = icl_list_new();
                    prt_assert(vsa->channel_lists[dst_node] != NULL,
                        "icl_list_new failed");
                }
                // Add the channel to the list.
                icl_list_t *node = icl_list_isort(
                    vsa->channel_lists[dst_node], channel, prt_channel_compare);
                prt_assert(node != NULL, "icl_list_isort failed");
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Assign channel tags.
 * Build the node-tag lookup.
 * Destroy channel lists.
 *
 * @param vsa
 */
void prt_vsa_channel_tags(prt_vsa_t *vsa)
{
    int i;
    for (i = 0; i < vsa->num_nodes; i++)
        if (vsa->channel_lists[i] != NULL) {
            int tag = 0;
            icl_list_t *node;
            // Assign consecutive tags to the elements.
            icl_list_foreach(vsa->channel_lists[i], node) {
                prt_channel_t *channel = (prt_channel_t*)node->data;
                channel->tag = tag++;

                int *node_tag;
                if (channel->dst_node == vsa->node_rank)
                    node_tag = prt_tuple_new2(channel->src_node, channel->tag);
                else
                    node_tag = prt_tuple_new2(channel->dst_node, channel->tag);

                icl_entry_t *entry = icl_hash_insert(
                    vsa->proxy->tags_hash, (void*)node_tag, (void*)channel);
                prt_assert(entry != NULL, "icl_hash_insert failed");
            }
            // Destroy the list.
            int status = icl_list_destroy(vsa->channel_lists[i], NULL);
            prt_assert(status == 0, "icl_list_destroy failed");
        }
    // Free the array of lists.
    free(vsa->channel_lists);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API
 * @brief VSA's production cycle
 * Launches worker threads.
 * Sends the master thread in the communication proxy production cycle.
 * Joins the worker threads.
 *
 * @param vsa
 */
void prt_vsa_run(prt_vsa_t *vsa)
{
    // Assign channel tags.
    prt_vsa_channel_tags(vsa);

    // Init tracing.
    svg_trace_init(vsa->num_threads);

    int i;
    // Launch threads.
    for (i = 0; i < vsa->num_threads; i++) {
        int status =
            pthread_create(
                &vsa->thread[i]->id, &vsa->thread_attr, prt_thread_run, vsa->thread[i]);
        prt_assert(status == 0, "pthread_create failed");
    }
    // Barrier for trace alignment.
    svg_trace_start(0);
    MPI_Barrier(MPI_COMM_WORLD);
    svg_trace_stop(0, Indigo);

    // Service the communication proxy.
    prt_proxy_run(vsa->proxy);

    // Join threads.
    for (i = 0; i < vsa->num_threads; i++) {
        int status = pthread_join(vsa->thread[i]->id, NULL);
        prt_assert(status == 0, "pthread_join failed");
    }
    // Finish tracing.
    if (vsa->config->svg_tracing == PRT_SVG_TRACING_ON)
        svg_trace_finish();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API
 * @brief Set VSA configuration parameter.
 *
 * @param vsa
 * @param param
 * @param value
 */
void prt_vsa_config_set(
    prt_vsa_t *vsa, prt_config_param_t param, prt_config_value_t value)
{
    switch (param) {
        case PRT_VDP_SCHEDULING:
            switch (value) {
                case PRT_VDP_SCHEDULING_LAZY:
                case PRT_VDP_SCHEDULING_AGGRESSIVE:
                    vsa->config->vdp_scheduling = value;
                    break;
                default:
                    prt_error("invalid value");
                    break;
            }
            break;
        case PRT_SVG_TRACING:
            switch (value) {
                case PRT_SVG_TRACING_ON:
                case PRT_SVG_TRACING_OFF:
                    vsa->config->svg_tracing = value;
                    break;
                default:
                    prt_error("invalid value");
                    break;
            }
            break;
        default:
            prt_error("invalid param");
            break;
    }
}
