/**
 * @file prt_vsa.h
 * @brief Virtual Systolic Array (VSA)
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.eecs.utk.edu/pulsar/
 * Copyright (C) 2012-2013 University of Tennessee.
 */

#ifndef PRT_VSA_H
#define PRT_VSA_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_proxy_s;
struct prt_config_s;
struct prt_thread_s;
enum prt_config_param_e;
enum prt_config_value_e;

/**
 * @brief maximum VDPs per node
 * She size of the VSA's hash table of VDPs.
 * Should be a prime number
 */
#define PRT_VSA_MAX_VDPS_PER_NODE 10003

/**
 * @brief VSA's VDPs mapping function pointer
 * Defines the pointer to the function mapping VDPs to cores.
 */
typedef int (*prt_vdp_map_func_t)(int*, void*, int);

/**
 * @brief Virtual Systolic Array (VSA)
 * VSA contains global informationa about the system,
 * a local communication proxy and an array of local worker threads.
 */
typedef struct prt_vsa_s {
    int node_rank;                  // node ID
    int num_nodes;                  // total number of nodes
    int num_threads;                // number of local threads
    int num_cores;                  // total number of worker cores
    pthread_attr_t thread_attr;     // attributes of worker threads
    struct prt_thread_s **thread;   // array of worker threads
    void *global_store;             // VSA's global store accessible by VDPs
    prt_vdp_map_func_t vdp_to_core; // function mapping VDPs' tuples to workers
    icl_hash_t *vdps_hash;          // VDPs lookup by tuple
    struct prt_config_s *config;    // VSA's configuration parameters
    struct prt_proxy_s *proxy;      // VSA's communication proxy
    icl_list_t **channel_lists;     // list of channels to each node
} prt_vsa_t;

prt_vsa_t* prt_vsa_new(
    int num_threads, void *global_store, int (*vdp_to_core)(int*, void*, int));
void prt_vsa_delete(prt_vsa_t *vsa);
void prt_vsa_vdp_insert(prt_vsa_t *vsa, struct prt_vdp_s *vdp);
void prt_vsa_vdp_merge_channels(
    prt_vsa_t *vsa, struct prt_vdp_s *vdp, int node_rank);
void prt_vsa_vdp_track_tags(
    prt_vsa_t *vsa, struct prt_vdp_s *vdp, int core, int node_rank);
void prt_vsa_channel_tags(prt_vsa_t *vsa);
void prt_vsa_run(prt_vsa_t *vsa);
void prt_vsa_config_set(
    prt_vsa_t *vsa, enum prt_config_param_e param, enum prt_config_value_e value);

#ifdef __cplusplus
}
#endif

#endif /* PRT_VSA_H */
