
#include <prt.h>
#include <plasma.h>
#include "pulsar_util.h"

/**
 * @brief LU nopiv VSA's global store
 */
typedef struct {
    int ib;
    int nb;
    int mt;
    int nt;
    int check;
} lu_nopiv_global_store_t;

/**
 * @brief LU nopiv VDP's local store
 */
typedef struct {
    prt_packet_t *L;
    prt_packet_t *U;
    prt_packet_t *A;
} lu_nopiv_local_store_t;

int lu_nopiv_vdp_to_core(int *tuple, void *global_store, int num_cores);
void lu_nopiv_vsa_init(prt_vsa_t *vsa, int ib, int nb, int mt, int nt);
void lu_nopiv_vdp_function1(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store);
void lu_nopiv_vdp_function2(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store);
void plasma_lu_nopiv(int nb, int ib, int mt, int nt, int num_threads);

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief main
 * Set up and run the VSA for the LU nopiv factorization (multi-node, multi-core).
 * Optionally call PLASMA LU nopiv factorization (single-node, multi-core).
 * Print corner tiles (bottom-right) for comparison.
 * Measure and report performance for each case.
 *
 * @param argc
 * @param argv
 *
 * @retval 0 success
 * @retval EXIT_FAILURE error
 */
int main(int argc, char **argv)
{
    if (argc < 5) {
        printf("usage: mpirun -n <proc> ./tile_qr "
               "<ib> <nb> <tiles> <threads> [test] [trace]\n"
               "  proc     number of MPI processes\n"
               "  ib       inner blocking, e.g., 32, 64, ...\n"
               "  nb       tile size, e.g., 192, 256, ...\n"
               "  tiles    number of tiles in each row and each column\n"
               "  threads  number of threads per process (use cores-1)\n"
               "  test     compare against PLASMA on one node\n\n"
               "  trace    print trace to an SVG file\n\n");
        exit(EXIT_FAILURE);
    }
    int iam;
    int ib = atoi(argv[1]);
    int nb = atoi(argv[2]);
    int bb = atoi(argv[3]);
    int num_threads = atoi(argv[4]);
    int mt = bb;
    int nt = bb;
    int m = nb*mt;
    int n = nb*nt;

    int test = 0;
    int trace = 0;
    if (argc == 6) {
        test  = !strcmp(argv[5], "test");
        trace = !strcmp(argv[5], "trace");
    }
    else if (argc == 7) {
        test  = !strcmp(argv[5], "test")  || !strcmp(argv[6], "test");
        trace = !strcmp(argv[5], "trace") || !strcmp(argv[6], "trace");
    }
    int mpi_rank;
    int mpi_size;
    assert(MPI_Init(&argc, &argv) == MPI_SUCCESS);
    assert(MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank) == MPI_SUCCESS);
    assert(MPI_Comm_size(MPI_COMM_WORLD, &mpi_size) == MPI_SUCCESS);
    if (mpi_rank == 0)
    printf( "\n > tile_lu_nopiv ib=%d nb=%d mt=nt=%d num_threads=%d\n",ib,nb,bb,num_threads );

    // Set Pulsar Global variables
    lu_nopiv_global_store_t lu_nopiv_global_store;
    lu_nopiv_global_store.ib = ib;
    lu_nopiv_global_store.nb = nb;
    lu_nopiv_global_store.mt = mt;
    lu_nopiv_global_store.nt = nt;
    lu_nopiv_global_store.check = test;


    // Create and Initialize VSA
    prt_vsa_t *vsa = prt_vsa_new(num_threads, &lu_nopiv_global_store, lu_nopiv_vdp_to_core);
    //prt_vsa_config_set(vsa, PRT_VDP_SCHEDULING, PRT_VDP_SCHEDULING_AGGRESSIVE);
    prt_vsa_config_set(vsa, PRT_VDP_SCHEDULING, PRT_VDP_SCHEDULING_LAZY);
    if (trace)
        prt_vsa_config_set(vsa, PRT_SVG_TRACING, PRT_SVG_TRACING_ON);

    // Run VSA
    double start, stop;
    start = pulsar_time_of_day();
    lu_nopiv_vsa_init(vsa, ib, nb, mt, nt);
    prt_vsa_run(vsa);
    prt_vsa_delete(vsa);
    stop = pulsar_time_of_day();
    if (mpi_rank == 0) {
        double gops = ((double)m*n*n - ((double)n)*n*n/3.0 - ((double)n)*n/2.0 + 5.0*n/6.0)/1000000000.0;
        printf("\nPULSAR\n");
        printf("  nodes:   %d\n", mpi_size);
        printf("  cores:   %d\n", mpi_size*num_threads);
        printf("  seconds: %lf\n", stop-start);
        printf("  gflops:  %lf\n", gops/(stop-start));

        if (test) {
            start = pulsar_time_of_day();
            plasma_lu_nopiv(nb, ib, mt, nt, num_threads);
            stop = pulsar_time_of_day();
            printf("\nPLASMA\n");
            printf("  nodes:   %d\n", 1);
            printf("  cores:   %d\n", num_threads);
            printf("  seconds: %lf\n", stop-start);
            printf("  gflops:  %lf\n", gops/(stop-start));
        }
        printf("\n");
    }

    // Cleanup
    assert(MPI_Finalize() == MPI_SUCCESS);
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines z-curve mapping of VDPs to cores.
 * 0
 * 1 2
 * 2 3 3
 * 3 0 0 3
 * 0 1 1 0 2
 * 1 2 2 1 3 0
 *
 * @param tuple VDP's tuple
 * @param global_store VSA's global store
 * @param num_cores VSA's (total) number of cores
 *
 * @return VDP's core (global) number
 */
int lu_nopiv_vdp_to_core(int *tuple, void *global_store, int num_cores)
{
    lu_nopiv_global_store_t *lu_nopiv_global_store = (lu_nopiv_global_store_t*)global_store;
    int mt = lu_nopiv_global_store->mt;
    int nt = lu_nopiv_global_store->nt;
    int m = tuple[0];
    int n = tuple[1];
    return ((m-n + n*(mt+mt-n+1)/2) % num_cores);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief VSA construction
 * Create the VDPs.
 * Create the channels.
 * Insert the channels in the VDPs.
 * Insert the VDPs in the VSA.
 *
 * @param vsa
 * @param nb matrix tile size
 * @param ib inner blocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 *
 */
void lu_nopiv_vsa_init(prt_vsa_t *vsa, int ib, int nb, int mt, int nt)
{
    int m;
    int n;
    prt_channel_t *channel;

    for (n = 0; n < nt; n++) {
        for (m = n; m < mt; m++) {
            prt_vdp_t *vdp;

            // Create VDP.
            if (m == n) { // top row
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-n, lu_nopiv_vdp_function1, sizeof(lu_nopiv_local_store_t), 2, 2, Red);
            } else {     // remaining rows
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-n, lu_nopiv_vdp_function2, sizeof(lu_nopiv_local_store_t), 2, 2, Green);
            }

            /* === INPUT chanels === */
            // Input channel 0 (A).
            if (n > 0) { /* if( n==0 ) then allocate and init tile */
                         /* else take it from A(m,n-1)             */
                channel = prt_channel_new(
                    nb*nb, MPI_DOUBLE, prt_tuple_new2(m, n-1), 0, prt_tuple_new2(m, n), 0);
                prt_vdp_channel_insert(vdp, channel, PrtInputChannel, 0);
            }
            // Input channel 1 (U).
            if (m > n) { /* remaining rows: pass U through from (m-1,n) */
                channel = prt_channel_new(
                    nb*nb, MPI_DOUBLE, prt_tuple_new2(m-1, n), 1, prt_tuple_new2(m, n), 1);
                prt_vdp_channel_insert(vdp, channel, PrtInputChannel, 1);
            }

            /* === OUTPUT chanels === */
            // Output channel 0 (A).
            if (m > n) { /* remaining rows: pass A to right */
                channel = prt_channel_new(
                    nb*nb, MPI_DOUBLE, prt_tuple_new2(m, n), 0, prt_tuple_new2(m, n+1), 0);
                prt_vdp_channel_insert(vdp, channel, PrtOutputChannel, 0);
            }
            // Output channel 1 (U).
            if (m+1 < mt) { /* pass U through to (m+1,n) */
                channel = prt_channel_new(
                    nb*nb, MPI_DOUBLE, prt_tuple_new2(m, n), 1, prt_tuple_new2(m+1, n), 1);
                prt_vdp_channel_insert(vdp, channel, PrtOutputChannel, 1);
            }
            prt_vsa_vdp_insert(vsa, vdp);
      }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an edge VDP (m == n), top rows
 *
 * @param tuple VDP's tuple
 * @param counter VDP's counter
 * @param input array of VDP's input channels
 * @param output array of VDP's output channels
 * @param local_store VDP's local store
 * @param global_store VSA's global store
 *                 ___
 *  A(input 0) -->|___|
 *                  |
 *                  V
 *             U(output 1)
 */
void lu_nopiv_vdp_function1(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store)
{
    lu_nopiv_global_store_t *qs = (lu_nopiv_global_store_t*)global_store;
    lu_nopiv_local_store_t  *ls = (lu_nopiv_local_store_t *)local_store;

    int ib = qs->ib;
    int nb = qs->nb;
    int nt = qs->nt;
    int mt = qs->mt;
    int m = tuple[0];
    int n = tuple[1];

    if (counter == nt-n) {
        /* diagonal block: diagonal factorization */
        if (n == 0) {
            // First touch creates/initializes A.
            ls->L = prt_packet_new(nb*nb*sizeof(double));
            //ls->L->data = &(qs->A[(m+(nt-counter)*mt)*(nb*nb)]);
            pulsar_tile_init(ls->L->data, m, nt-counter, nb);
        }
        else {
            // Otherwise, pop A.
            ls->L = prt_channel_pop(input[0]);
        }
        // DGETRF_nopiv.
        CORE_dgetrf_nopiv(nb, nb, ib, ls->L->data, nb);
        if (m+1 < mt) { /* push U down */
            prt_channel_push(output[1], ls->L);
        }
    }
    else {
        /* remaining blocks in the top row:    *
        *   computing off-diagonal blocks of U */
        if (n == 0) {
            // First touch creates/initializes A.
            ls->U = prt_packet_new(nb*nb*sizeof(double));
            //ls->U->data = &(qs->A[(m+(nt-counter)*mt)*(nb*nb)]);
            pulsar_tile_init(ls->U->data, m, nt-counter, nb);
        }
        else {
            // Otherwise, pop A.
            ls->U = prt_channel_pop(input[0]);
        }
        // DTRSM.
        /* U(m,n) = L(m,m)^-1 A(m,n) */
        CORE_dtrsm( PlasmaLeft, PlasmaLower, PlasmaNoTrans, PlasmaUnit,
                    nb, nb, 1.0, ls->L->data, nb, ls->U->data, nb);
        /* push U down */
        prt_channel_push(output[1], ls->U);
    }

    // IF last step.
    if (counter == nt-n && qs->check == 1)
    {
        // IF the last tile - print.
        if (m == mt-1 && n == nt-1) {
            pulsar_tile_print((double*)ls->L->data, nb);
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an out-of-edge VDP (m != n), remaining rows
 *
 * @param tuple VDP's tuple
 * @param counter VDP's counter
 * @param input array of VDP's input channels
 * @param output array of VDP's output channels
 * @param local_store VDP's local store
 * @param global_store VSA's global store
 *
 *              U(input 1)
 *                  |
 *                 _V_
 *  A(input 0) -->|___|--> A(output 0)
 *                  |
 *                  V
 *             U(output 1) - pass-through
 */
void lu_nopiv_vdp_function2(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store)
{
    lu_nopiv_global_store_t *qs = (lu_nopiv_global_store_t*)global_store;
    lu_nopiv_local_store_t *ls = (lu_nopiv_local_store_t*)local_store;

    int ib = qs->ib;
    int nb = qs->nb;
    int nt = qs->nt;
    int mt = qs->mt;
    int m = tuple[0];
    int n = tuple[1];

    if (counter == nt-n) {
        /* first column, compute off-diagonal L (panel factorization) */
        if (n == 0) {
            // First touch creates/initializes A.
            ls->L = prt_packet_new(nb*nb*sizeof(double));
            //ls->L->data = &(qs->A[(m+(nt-counter)*mt)*(nb*nb)]);
            pulsar_tile_init(ls->L->data, m, nt-counter, nb);
        }
        else {
            // Otherwise, pop A.
            ls->L = prt_channel_pop(input[0]);
        }
        // pop U from (m-1,n)
        ls->U = prt_channel_pop(input[1]);
        if (m < mt-1) {
            // pass U down to (m+1,n)
            prt_channel_push(output[1], ls->U);
        }
        // DTRSM
        /* L(m,n) = A(m,n) U(n,n)^-1 */
        CORE_dtrsm(PlasmaRight, PlasmaUpper, PlasmaNoTrans, PlasmaNonUnit,
                   nb, nb, 1.0, ls->U->data, nb, ls->L->data, nb );
    }
    else {
        /* rest of columns, trailing submatrix update */
        if (n == 0) {
            // First touch creates/initializes A.
            ls->A = prt_packet_new(nb*nb*sizeof(double));
            //ls->A->data = &(qs->A[(m+(nt-counter)*mt)*(nb*nb)]);
            pulsar_tile_init(ls->A->data, m, nt-counter, nb);
        }
        else {
            // Otherwise, pop A.
            ls->A = prt_channel_pop(input[0]);
        }
        // pop U from (m-1,n)
        ls->U = prt_channel_pop(input[1]);
        if (m < mt-1) {
            // pass U down to (m+1,n)
            prt_channel_push(output[1], ls->U);
        }
        // GEMM
        CORE_dgemm(PlasmaNoTrans, PlasmaNoTrans, nb, nb, nb,
                   -1.0, ls->L->data, nb,
                         ls->U->data, nb,
                    1.0, ls->A->data, nb);
        // push A to right
        prt_channel_push(output[0], ls->A);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Call PLASMA LU nopiv factorization.
 * Make a single-node multi-core PLASMA LU nopiv run.
 * Set the parameters and initialize the matrix to match the VSA run.
 *
 * @param nb matrix tile size
 * @param ib inner bocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 * @param num_threads number of threads
 */
void plasma_lu_nopiv(int nb, int ib, int mt, int nt, int num_threads)
{
    PLASMA_Init(num_threads);
    PLASMA_desc *descA;
    PLASMA_Disable(PLASMA_AUTOTUNING);                                           // Initialize PLASMA
    PLASMA_Set(PLASMA_TILE_SIZE, nb);
    PLASMA_Set(PLASMA_INNER_BLOCK_SIZE, ib);
    int i,j;                                                                     // Initialize matrix
    double *A = (double*)malloc(mt*nt*nb*nb*sizeof(double));
    for (j = 0; j < mt; j++)
        for (i = 0; i < nt; i++)
            pulsar_tile_init(&A[(i*mt+j)*nb*nb], j, i, nb);
    PLASMA_Desc_Create(
        &descA, A, PlasmaRealDouble,
        nb, nb, nb*nb, mt*nb, nt*nb, 0, 0, mt*nb, nt*nb);
    PLASMA_dgetrf_nopiv_Tile(descA);                                             // Call PLASMA
    pulsar_tile_print(&A[((nt-1)*mt+(mt-1))*nb*nb], nb);
    PLASMA_Desc_Destroy(&descA);                                                 // Cleanup
    PLASMA_Finalize();
}
