
#include <prt.h>
#include <plasma.h>
#include "pulsar_util.h"

/**
 * @brief tile QR VSA's global store
 */
typedef struct {
    int ib;
    int nb;
    int mt;
    int nt;
} qr_global_store_t;

/**
 * @brief tile QR VDP's local store
 */
typedef struct {
    prt_packet_t *A1;
    prt_packet_t *A2;
    prt_packet_t *T;
    prt_packet_t *V;
    double *work;
    double *tau;
} qr_local_store_t;

int qr_vdp_to_core(int *tuple, void *global_store, int num_cores);
void qr_vsa_init(prt_vsa_t *vsa, int nb, int ib, int mt, int nt);
void qr_vdp_function1(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store);
void qr_vdp_function2(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store);
void plasma_tile_qr(int nb, int ib, int mt, int nt, int num_threads);

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief main
 * Set up and run the VSA for the tile QR factorization (multi-node, multi-core).
 * Optionally call PLASMA tile QR factorization (single-node, multi-core).
 * Print corner tiles (bottom-right) for comparison.
 * Measure and report performance for each case.
 *
 * @param argc
 * @param argv
 *
 * @retval 0 success
 * @retval EXIT_FAILURE error
 */
int main(int argc, char **argv)
{
    if (argc < 5) {
        printf("usage: mpirun -n <proc> ./tile_qr "
               "<ib> <nb> <tiles> <threads> [test] [trace]\n"
               "  proc     number of MPI processes\n"
               "  ib       inner blocking, e.g., 32, 64, ...\n"
               "  nb       tile size, e.g., 192, 256, ...\n"
               "  tiles    number of tiles in each row and each column\n"
               "  threads  number of threads per process (use cores-1)\n"
               "  test     compare against PLASMA on one node\n\n"
               "  trace    print trace to an SVG file\n\n");
        exit(EXIT_FAILURE);
    }
    int ib = atoi(argv[1]);
    int nb = atoi(argv[2]);
    int bb = atoi(argv[3]);
    int num_threads = atoi(argv[4]);
    int mt = bb;
    int nt = bb;

    int test;
    int trace;
    if (argc == 6) {
        test  = !strcmp(argv[5], "test");
        trace = !strcmp(argv[5], "trace");
    }
    else if (argc == 7) {
        test  = !strcmp(argv[5], "test")  || !strcmp(argv[6], "test");
        trace = !strcmp(argv[5], "trace") || !strcmp(argv[6], "trace");
    }
    qr_global_store_t qr_global_store;
    qr_global_store.ib = ib;
    qr_global_store.nb = nb;
    qr_global_store.mt = mt;
    qr_global_store.nt = nt;

    int mpi_rank;
    int mpi_size;
    assert(MPI_Init(&argc, &argv) == MPI_SUCCESS);
    assert(MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank) == MPI_SUCCESS);
    assert(MPI_Comm_size(MPI_COMM_WORLD, &mpi_size) == MPI_SUCCESS);

    prt_vsa_t *vsa = prt_vsa_new(num_threads, &qr_global_store, qr_vdp_to_core);
    prt_vsa_config_set(vsa, PRT_VDP_SCHEDULING, PRT_VDP_SCHEDULING_AGGRESSIVE);
//  prt_vsa_config_set(vsa, PRT_VDP_SCHEDULING, PRT_VDP_SCHEDULING_LAZY);
    if (trace)
        prt_vsa_config_set(vsa, PRT_SVG_TRACING, PRT_SVG_TRACING_ON);

    double start;
    double stop;
    start = pulsar_time_of_day();
    qr_vsa_init(vsa, nb, ib, mt, nt);
    prt_vsa_run(vsa);
    prt_vsa_delete(vsa);
    stop = pulsar_time_of_day();

    if (mpi_rank == 0) {
        double gops = 4.0/3.0*nb*nb*nb*bb*bb*bb/1000000000.0;
        printf("\nPULSAR\n");
        printf("  nodes:   %d\n", mpi_size);
        printf("  cores:   %d\n", mpi_size*num_threads);
        printf("  seconds: %lf\n", stop-start);
        printf("  gflops:  %lf\n", gops/(stop-start));

        if (test) {
            start = pulsar_time_of_day();
            plasma_tile_qr(nb, ib, mt, nt, num_threads);
            stop = pulsar_time_of_day();
            printf("\nPLASMA\n");
            printf("  nodes:   %d\n", 1);
            printf("  cores:   %d\n", num_threads);
            printf("  seconds: %lf\n", stop-start);
            printf("  gflops:  %lf\n", gops/(stop-start));
        }
        printf("\n");
    }
    assert(MPI_Finalize() == MPI_SUCCESS);
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines z-curve mapping of VDPs to cores.
 * 0 1 2 3 0 1
 *   2 3 0 1 2
 *     3 0 1 2
 *       3 0 1
 *         2 3
 *           0
 *
 * @param tuple VDP's tuple
 * @param global_store VSA's global store
 * @param num_cores VSA's (total) number of cores
 *
 * @return VDP's core (global) number
 */
int qr_vdp_to_core(int *tuple, void *global_store, int num_cores)
{
    qr_global_store_t *qr_global_store = (qr_global_store_t*)global_store;
    int mt = qr_global_store->mt;
    int nt = qr_global_store->nt;
    int m = tuple[0];
    int n = tuple[1];
    return ((mt*m+n - m*(m+1)/2) % num_cores);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief VSA construction
 * Create the VDPs.
 * Create the channels.
 * Insert the channels in the VDPs.
 * Insert the VDPs in the VSA.
 *
 * @param vsa
 @ @param nb matrix tile size
 * @param ib QR inner blocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 */
void qr_vsa_init(prt_vsa_t *vsa, int nb, int ib, int mt, int nt)
{
    int m;
    int n;
    prt_channel_t *channel;

    for (m = 0; m < mt; m++) {
        for (n = m; n < nt; n++) {
            prt_vdp_t *vdp; // Create the VDP.
            if (m == n)
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-m, qr_vdp_function1,
                    sizeof(qr_local_store_t), 3, 3, SteelBlue);
            else
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-m, qr_vdp_function2,
                    sizeof(qr_local_store_t), 3, 3, LightSteelBlue);

            if (m > 0) {    // input channel 0
                channel = prt_channel_new(nb*nb, MPI_DOUBLE,
                    prt_tuple_new2(m-1, n), 0, prt_tuple_new2(m, n), 0);
                prt_vdp_channel_insert(vdp, channel, PrtInputChannel, 0);
            }
            if (n > m) {    // input channel 1
                channel = prt_channel_new(nb*nb, MPI_DOUBLE,
                    prt_tuple_new2(m, n-1), 1, prt_tuple_new2(m, n), 1);
                prt_vdp_channel_insert(vdp, channel, PrtInputChannel, 1);
            }
            if (n > m) {    // input channel 2
                channel = prt_channel_new(ib*nb, MPI_DOUBLE,
                    prt_tuple_new2(m, n-1), 2, prt_tuple_new2(m, n), 2);
                prt_vdp_channel_insert(vdp, channel, PrtInputChannel, 2);
            }
            if (m < n) {    // output channel 0
                channel = prt_channel_new(nb*nb, MPI_DOUBLE,
                    prt_tuple_new2(m, n), 0, prt_tuple_new2(m+1, n), 0);
                prt_vdp_channel_insert(vdp, channel, PrtOutputChannel, 0);
            }
            if (n < nt-1) { // output channel 1
                channel = prt_channel_new(nb*nb, MPI_DOUBLE,
                    prt_tuple_new2(m, n), 1, prt_tuple_new2(m, n+1), 1);
                prt_vdp_channel_insert(vdp, channel, PrtOutputChannel, 1);
            }
            if (n < nt-1) { // output channel 2
                channel = prt_channel_new(ib*nb, MPI_DOUBLE,
                    prt_tuple_new2(m, n), 2, prt_tuple_new2(m, n+1), 2);
                prt_vdp_channel_insert(vdp, channel, PrtOutputChannel, 2);
            }
            // Insert the VDP.
            prt_vsa_vdp_insert(vsa, vdp);
      }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an edge VDP.
 * A(input 0)
 *     |
 *    _V_
 *   |   |--> V(output 1)
 *   |___|--> T(output 2)
 *
 * @param tuple VDP's tuple
 * @param counter VDP's counter
 * @param input array of VDP's input channels
 * @param output array of VDP's output channels
 * @param local_store VDP's local store
 * @param global_store VSA's global store
 */
void qr_vdp_function1(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store)
{
    qr_global_store_t *qs = (qr_global_store_t*)global_store;
    qr_local_store_t *ls = (qr_local_store_t*)local_store;

    int ib = qs->ib;
    int nb = qs->nb;
    int nt = qs->nt;
    int mt = qs->mt;
    int m = tuple[0];
    int n = tuple[1];

    if (counter == nt-m) {                                  // IF first step.
        ls->work = (double*)malloc(ib*nb*sizeof(double));       // Allocate workspace.
        ls->tau  = (double*)malloc(   nb*sizeof(double));
        assert(ls->work != NULL);
        assert(ls->tau  != NULL);

        if (m == 0) {                                           // First row creates A1.
            ls->A1 = prt_packet_new(nb*nb*sizeof(double));
            pulsar_tile_init(ls->A1->data, nt-counter, n, nb);
        }
        else {                                                  // Lower rows pop A1.
            ls->A1 = prt_channel_pop(input[0]);
        }
        ls->T = prt_packet_new(ib*nb*sizeof(double));
        CORE_dgeqrt(                                            // Call DGEQRT.
            nb, nb, ib,
            ls->A1->data, nb,
            ls->T->data, ib,
            ls->tau, ls->work);
        if (n < nt-1) {
            prt_channel_push(output[1], ls->A1);
            prt_channel_push(output[2], ls->T);
        }
        prt_packet_release(ls->T);
    }
    else {                                                  // Follow-up steps.
        if (m == 0) {                                           // First row creates A2.
            ls->A2 = prt_packet_new(nb*nb*sizeof(double));
            pulsar_tile_init(ls->A2->data, nt-counter, n, nb);
        }
        else {                                                  // Lower rows pop A2.
            ls->A2 = prt_channel_pop(input[0]);
        }
        ls->T = prt_packet_new(ib*nb*sizeof(double));
        CORE_dtsqrt(                                            // Call DTSQRT.
            nb, nb, ib,
            ls->A1->data, nb,
            ls->A2->data, nb,
            ls->T->data, ib,
            ls->tau, ls->work);
        prt_channel_push(output[1], ls->A2);
        prt_channel_push(output[2], ls->T);
        prt_packet_release(ls->A2);
        prt_packet_release(ls->T);
    }
    if (counter == 1) {                                     // IF last step.
        if (m == mt-1 && n == nt-1)                             // IF corner VDP.
            pulsar_tile_print((double*)ls->A1->data, nb);           // Print corner tile.
        prt_packet_release(ls->A1);                             // Release A1.
        free(ls->work);                                         // Free workspace.
        free(ls->tau);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an out-of-edge VDP.
 *             A(input 0)
 *                 |
 *                _V_
 * V(input 1) -->|   |--> V(output 1) - pass-through
 * T(input 2) -->|___|--> T(output 2) - pass-through
 *                 |
 *                 V
 *            A(output 0)
 *
 * @param tuple VDP's tuple
 * @param counter VDP's counter
 * @param input array of VDP's input channels
 * @param output array of VDP's output channels
 * @param local_store VDP's local store
 * @param global_store VSA's global store
 */
void qr_vdp_function2(
    int *tuple, int counter,
    prt_channel_t **input, prt_channel_t **output,
    void *local_store, void *global_store)
{
    qr_global_store_t *qs = (qr_global_store_t*)global_store;
    qr_local_store_t *ls = (qr_local_store_t*)local_store;

    int ib = qs->ib;
    int nb = qs->nb;
    int nt = qs->nt;
    int mt = qs->mt;
    int m = tuple[0];
    int n = tuple[1];

    if (counter == nt-m) {                                  // IF first step.
        ls->work = (double*)malloc(ib*nb*sizeof(double));       // Allocate workspace.
        ls->tau  = (double*)malloc(   nb*sizeof(double));
        assert(ls->work != NULL);
        assert(ls->tau  != NULL);

        if (m == 0) {                                           // First row creates A1.
            ls->A1 = prt_packet_new(nb*nb*sizeof(double));
            pulsar_tile_init(ls->A1->data, nt-counter, n, nb);
        }
        else {                                                  // Lower rows pop A1.
            ls->A1 = prt_channel_pop(input[0]);
        }
        ls->V = prt_channel_pop(input[1]);
        ls->T = prt_channel_pop(input[2]);
        if (n < nt-1) {
            prt_channel_push(output[1], ls->V);
            prt_channel_push(output[2], ls->T);
        }
        CORE_dormqr(                                            // Call DORMQR.
            PlasmaLeft, PlasmaTrans,
            nb, nb, nb, ib,
            ls->V->data, nb,
            ls->T->data, ib,
            ls->A1->data, nb,
            ls->work, nb);
        prt_packet_release(ls->V);
        prt_packet_release(ls->T);
    }
    else {                                                  // Follow-up steps.
        if (m == 0) {                                           // First row creates A2.
            ls->A2 = prt_packet_new(nb*nb*sizeof(double));
            pulsar_tile_init(ls->A2->data, nt-counter, n, nb);
        }
        else {                                                  // Lower rows pop A2.
            ls->A2 = prt_channel_pop(input[0]);
        }
        ls->V = prt_channel_pop(input[1]);
        ls->T = prt_channel_pop(input[2]);
        if (n < nt-1) {
            prt_channel_push(output[1], ls->V);
            prt_channel_push(output[2], ls->T);
        }
        CORE_dtsmqr(                                            // Call DTSMQR.
            PlasmaLeft, PlasmaTrans,
            nb, nb, nb, nb, nb, ib,
            ls->A1->data, nb,
            ls->A2->data, nb,
            ls->V->data, nb,
            ls->T->data, ib,
            ls->work, ib);
        prt_channel_push(output[0], ls->A2);
        prt_packet_release(ls->V);
        prt_packet_release(ls->T);
        prt_packet_release(ls->A2);
    }
    if (counter == 1) {                                     // IF last step.
        prt_packet_release(ls->A1);                             // Release A1.
        free(ls->work);                                         // Free workspace.
        free(ls->tau);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Call PLASMA tile QR factorization.
 * Make a single-node multi-core PLASMA tile QR run.
 * Set the parameters and initialize the matrix to match the VSA run.
 *
 * @param nb matrix tile size
 * @param ib QR inner bocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 * @param num_threads number of threads
 */
void plasma_tile_qr(int nb, int ib, int mt, int nt, int num_threads)
{
    PLASMA_Init(num_threads);
    PLASMA_desc *descA;
    PLASMA_desc *descT;
    PLASMA_Disable(PLASMA_AUTOTUNING);
    PLASMA_Set(PLASMA_TILE_SIZE, nb);
    PLASMA_Set(PLASMA_INNER_BLOCK_SIZE, ib);
    double *A = (double*)malloc(mt*nt*nb*nb*sizeof(double));
    int i;
    int j;
    for (j = 0; j < mt; j++)
        for (i = 0; i < nt; i++)
            pulsar_tile_init(&A[(i*mt+j)*nb*nb], j, i, nb);
    PLASMA_Desc_Create(
        &descA, A, PlasmaRealDouble,
        nb, nb, nb*nb, mt*nb, nt*nb, 0, 0, mt*nb, nt*nb);
    PLASMA_Alloc_Workspace_dgeqrf_Tile(mt*nb, nt*nb, &descT);
    PLASMA_dgeqrf_Tile(descA, descT);
    pulsar_tile_print(&A[((nt-1)*mt+(mt-1))*nb*nb], nb);
    PLASMA_Desc_Destroy(&descA);
    free(A);
    PLASMA_Finalize();
}
